@@ -24,13 +24,6 @@
 #include "linc-debug.h"
 #include "linc-private.h"
 
-struct _LINCServerPrivate {
-	int        fd;
-	GMutex    *mutex;
-	LincWatch *tag;
-	GSList    *connections;
-};
-
 enum {
 	NEW_CONNECTION,
 	LAST_SIGNAL
@@ -86,7 +79,7 @@ linc_server_dispose (GObject *obj)
 	GSList     *l;
 	LINCServer *cnx = (LINCServer *)obj;
 
-	d_printf ("Dispose / close server fd %d", cnx->priv->fd);
+	d_printf ("Dispose / close server fd %d\n", cnx->priv->fd);
 
 #ifdef G_THREADS_ENABLED
 	if (cnx->priv->mutex) {
@@ -221,6 +214,9 @@ linc_server_handle_io (GIOChannel  *gioc,
 		g_value_set_object (parms, G_OBJECT (server));
 		g_value_init (parms + 1, G_TYPE_OBJECT);
 		g_value_set_object (parms + 1, G_OBJECT (connection));
+
+		d_printf ("p %d, Non-accepted input on fd %d",
+			  getpid (), server->priv->fd);
 		
 		g_signal_emitv (parms, server_signals [NEW_CONNECTION], 0, NULL);
 		
@@ -239,11 +235,11 @@ linc_server_setup (LINCServer            *cnx,
 		   LINCConnectionOptions  create_options)
 {
 	const LINCProtocolInfo *proto;
-	GIOChannel             *gioc;
 	int                     fd, n;
 	struct sockaddr        *saddr;
 	socklen_t               saddr_len;
 	char                    hnbuf[NI_MAXHOST];
+	const char             *local_host;
 	char                   *service, *hostname;
 
 #if !LINC_SSL_SUPPORT
@@ -257,21 +253,24 @@ linc_server_setup (LINCServer            *cnx,
 		return FALSE;
 	}
 
-	if (!local_host_info) {
-		local_host_info = hnbuf;
-		if (gethostname (hnbuf, sizeof (hnbuf)))
+	if (local_host_info)
+		local_host = local_host_info;
+	else {
+		local_host = hnbuf;
+		if (gethostname (hnbuf, sizeof (hnbuf))) {
+			hnbuf[0] = '\0';
 			perror ("gethostname failed!");
+		}
 	}
 
  address_in_use:
 
-	saddr = linc_protocol_get_sockaddr (proto, local_host_info, 
-					    local_serv_info, &saddr_len);
+	saddr = linc_protocol_get_sockaddr (
+		proto, local_host, local_serv_info, &saddr_len);
 
 	if (!saddr) {
 		d_printf ("Can't get_sockaddr proto '%s' '%s'\n",
-			 local_host_info,
-			 local_serv_info ? local_serv_info : "(null)");
+			  local_host, local_serv_info ? local_serv_info : "(null)");
 		return FALSE;
 	}
 
@@ -308,10 +307,15 @@ linc_server_setup (LINCServer            *cnx,
 		d_printf ("bind really failed errno: %d\n", errno);
 
 	if (!n)
-		n = getsockname (fd, saddr, &saddr_len);
+		n = fcntl (fd, F_SETFD, FD_CLOEXEC);
 	else
 		d_printf ("listen failed errno: %d\n", errno);
 
+	if (!n)
+		n = getsockname (fd, saddr, &saddr_len);
+	else
+		d_printf ("failed to set cloexec on %d", fd);
+
 	if (n) {
 		close (fd);
 		d_printf ("get_sockname failed errno: %d\n", errno);
@@ -330,20 +334,22 @@ linc_server_setup (LINCServer            *cnx,
 	cnx->proto = proto;
 	cnx->priv->fd = fd;
 
-	if ((create_options & LINC_CONNECTION_NONBLOCKING)) {
-		gioc = g_io_channel_unix_new (fd);
-
+	if (create_options & LINC_CONNECTION_NONBLOCKING) {
 		g_assert (cnx->priv->tag == NULL);
 
-		cnx->priv->tag = linc_io_add_watch (
-			gioc, LINC_IN_CONDS | LINC_ERR_CONDS,
+		cnx->priv->tag = linc_io_add_watch_fd (
+			fd, LINC_IN_CONDS | LINC_ERR_CONDS,
 			linc_server_handle_io, cnx);
-
-		g_io_channel_unref (gioc);
 	}
 
-	cnx->create_options  = create_options;
-	cnx->local_host_info = hostname;
+	cnx->create_options = create_options;
+
+	if (local_host_info) {
+		g_free (hostname);
+		cnx->local_host_info = g_strdup (local_host_info);
+	} else
+		cnx->local_host_info = hostname;
+
 	cnx->local_serv_info = service;
 
 	d_printf ("Created a new server fd (%d) '%s', '%s', '%s'\n",
